//// データ関係のクラス
// 結合されたデータのバッファー
class Data {
  OneData _top, _now;    // リスト構造の先頭と現在の書き込み・読み込みポインター
  int     num;           // 読み込まれたファイル数
  // コンストラクター
  Data() {
    _top = new OneData();
    gotoTop();
    num  = 0;
  }
  // リストの先頭に戻る
  void gotoTop() {
    _now = _top;
  }
  // リストの次の要素に移動する
  OneData next() {
    if (_now.next == null) return null;
    _now = _now.next;
    return _now;
  }
  // リストを全て削除する
  // （再起呼び出しを使った方がコードはシンプルだが…）
  void allclear() {
    if (_top.next == null) return;
    println("Clear all data");
    gotoTop();
    while (_now.next != null) {
      OneData next;

      next       = _now.next;
      _now.next  = null;
      _now       = next;
    }
    _top = new OneData();
    gotoTop();
    num  = 0;
  }
  // １ファイルのデータをメモリにストア（リストを追加）する
  OneData store(OneData onedata) {
    _now.next = _now.add(onedata);
    _now      = _now.next;
    num++;
    return _now;
  }
}

// １ファイル分のデータ
class OneData {  // リスト構造
  String filename;
  int    org_width, org_height;
  float  scale;
  int    num;
  int [] px    = new int[Max_measure];
  int [] py    = new int[Max_measure];
  int [] col_R = new int[Max_measure];
  int [] col_G = new int[Max_measure];
  int [] col_B = new int[Max_measure];
  OneData next;
  
  // コンストラクター
  OneData() {
    filename   = "";
    org_width  = 0; 
    org_height = 0;
    scale      = 0.0;
    next       = null;
  }
  // データを書き込んで次のリストを作成する
  OneData add(OneData onedata) {
    int i;
    filename   = onedata.filename;
    org_width  = onedata.org_width;
    org_height = onedata.org_height;
    scale      = onedata.scale;
    num        = onedata.num;
    for (i = 0; i < num; i++) {
      px[i]    = onedata.px[i];
      py[i]    = onedata.py[i];
      col_R[i] = onedata.col_R[i];
      col_G[i] = onedata.col_G[i];
      col_B[i] = onedata.col_G[i];
    }
    next = new OneData();
    return next;
  }
}

//// ファイル関係のクラス
//// ファイル関係のクラス
// CSV（タブ区切りの整数）のテキストを整数値に切り分けるクラス
class CsvStrings {
  String text;
  String str;
  float  value;
  // コンストラクター
  CsvStrings(String text) {
    this.set(text);
  }
  void set(String text) {
    this.text = text + "\t";
    value     =  float("NaN");  // floatの形式ではないのでNaNになる
  }
  // 次の値を取得する：正常に終了したらtrueを返し，valueに値が設定される
  boolean next() {
    if (text.indexOf("\t") == -1) return false;
    if (text.length() == 0)       return false;
    if (text.equals("\t"))        return false;
    str   = text.substring(0, text.indexOf("\t"));
    value = float(str);
    text  = text.substring(text.indexOf("\t") + 1);
    return true;
  }
  boolean is_float() {
    if (value != value) return false;  // NaN, ちょっとトリッキーな方法
    return true;
  }
}

// 指定されたデータファイルから１画像分のデータを取得してメモリーに保存するクラス
class ImportFile {
  // コンストラクター
  ImportFile() {
  }

  // ファイル形式のチェック
  boolean validation(String fname) {
    File         fp;
    String[]     lines;  // データファイルから読み出した全テキストデータ

    // 読み込めるか？
    if ((lines = loadStrings(fname)) == null) return false;
    // 少なくとも２行以上か？
    if ((lines.length < 2)) return false;
    // １行目はscale, org_width, org_heightか？
    CsvStrings csv = new CsvStrings(lines[0]);
    csv.next();
    if (! csv.is_float() || csv.value <= 0.0)     return false;  // scale (float)
    csv.next();
    if (! csv.is_float() || int(csv.value) <= 0)  return false;  // org_width (int)
    csv.next();
    if (! csv.is_float() || int(csv.value) <= 0)  return false;  // org_height (int)
    // 互換性（Ver. 1.*のGokuraku_measureは画像ファイル名を記録していない）ので，この後に文字列があるかどうかはチェックしない
    // ２行目は0で始まるか？
    csv.set(lines[1]);
    csv.next();
    if (! csv.is_float() || int(csv.value) != 0)  return false;

    return true;
}
  
  // 実際にファイルのデータを読み込む
  boolean read(String fname, Data data) {
    File         fp;
    CsvStrings   csv;
    OneData      onedata, now;
    String[]     lines;

    // ファイルチェック
    if (! validation(fname)) return false; 
    // ファイル読み込み
    if ((lines = loadStrings(fname)) == null) return false;

    // Scale, org_width, org_heightの読み出し
    onedata = new OneData();
    csv     = new CsvStrings(lines[0]);
    csv.next();
    onedata.scale      = csv.value;
    csv.next();
    onedata.org_width  = int(csv.value);
    csv.next();
    onedata.org_height = int(csv.value);
    csv.next();  // Ver.2 から，Gokuraku_maesureで画像ファイル名を記録するように変更
    if (csv.str.length() > 0) {
      onedata.filename      = csv.str;
    } else onedata.filename = "";

    // データ本体の読み込み
    for (int i = 1; i < lines.length; i++) {
      int value;
      csv.set(lines[i]);
      csv.next();
      if (int(csv.value) != (i - 1)) {
        shortMessage.set_message("Error: データ形式が正しくない（No.が一致しない）");
        return false;
      }
      // X座標
      csv.next();
      if (Scale_convert) onedata.px[i - 1] = int(csv.value / onedata.scale);
      else               onedata.px[i - 1] = int(csv.value);
      // Y座標
      csv.next();
      if (Scale_convert) onedata.py[i - 1] = int(csv.value / onedata.scale);
      else               onedata.py[i - 1] = int(csv.value);
      // R, G の空読み
      csv.next();
      onedata.col_R[i - 1] = int(csv.value);
      csv.next();
      onedata.col_G[i - 1] = int(csv.value);
      // B が存在するかどうかで一行のデータ数を確認
      if (csv.next() == false) {
        shortMessage.set_message("Error: データ形式が正しくない（一行のデータ数が少ない)");
        return false;
      }
      onedata.col_B[i - 1] = int(csv.value);
      onedata.num      = i;  // 実際に読み込んだ行数（更新する）
    }
    if (onedata.filename.length() == 0) {
      fp = new File(fname);  // ファイル名を分離するために．
      onedata.filename = fp.getName();
    }
    // データの格納
    now = data._now;
    data.store(onedata);
    shortMessage.set_message("Data imported: " + now.filename + " " + now.num + " points");
    if (now.next == null) {
      shortMessage.set_message("Error: データのメモリへの格納に失敗しました");
      return false;
    }
    return true;
  }     
}

// 結合したデータをファイルに出力するクラス
class ExportFile {
  String      export_folder;
  String      fullpathname;
  String      dest_filename;
  PrintWriter output;
  
  // コンストラクター
  ExportFile() {
    dest_filename = DataFilename;
    export_folder = sketchPath();
    set_export_folder(export_folder);
  }
  void set_export_folder(String path) {
    if (path.equals("")) export_folder = sketchPath();  // この指定は無いはず
    else                 export_folder = path;
    File f = new File(export_folder, dest_filename);
    fullpathname = f.getAbsolutePath(); 
  }
  // 結合したデータのファイル出力
  boolean export_alldata(Data data) {
    OneData now;
    if (data._top.next == null) return false;  // まだデータが無い

    output = createWriter(fullpathname);
    if (output == null) {
      shortMessage.set_message("Error: 出力ファイルが開けません : " + fullpathname);
      return false;
    } else println("Start exporting to " + fullpathname);

    // データファイルの出力
    int num = 0;;
    now = data._top;
    while (true) {
      // ヘッダーの書き込み
      output.print(now.filename);                                   // ファイル名
      output.print("\t" + now.scale);                               // 倍率（Scale_convertがtrueの場合はこの倍率でX, Y座標は変換されて結合される）
      output.print("\t" + now.org_width + "\t" + now.org_height);   // 画像の幅，高さ
      output.print("\t" + now.num);                                 // 計測点数
      // データ本体の書き込み
      for (int i = 0; i < now.num; i++) {
        output.print("\t" + now.px[i] + "\t" + now.py[i]);                                // X, Y座標
        // 色情報も出力する必要があるならば
        if (Output_color) {
          output.print("\t" + now.col_R[i] + "\t" + now.col_G[i] + "\t" + now.col_B[i]);  // R, G, B
        }
      }
      // 最後の改行
      output.println();
      num++;
      now = now.next;
      if (now.next == null) break;
    }    
    output.flush();
    output.close();
    shortMessage.set_message("Completde export " + num + " data files");
    data.allclear();
    return true;
  }
}

// ウインドウとコンソールエリアに短いメッセージを一時的に表示するためのクラス
class ShortMessage {
private  int    start;
private  String message;
  ShortMessage() {
    start   = millis();
    message = "";
  }
  void set_message(String str) {
    message = str;
    println(str);
    start   = millis();
//  delay(500);          // 連続表示に対応するために．ただし動作が遅くなるため無効にしておくのでご自由に．
  }
  void display_short_message(int x, int y) {
    if (millis() - start < Delay_message) {
      fill(255, 255, 255);
      textSize(16);
      text(message, x, y);
    }
  }
}
